import pytest
from mock import Mock


@pytest.fixture
def meth():
    from twitterdedupe import lengthen_url
    return lengthen_url


def reqlib(url):
    reqlib = Mock(name="requests")
    response = Mock(name="response")
    response.url = url
    reqlib.get.return_value = response

    r = reqlib.get(url)
    assert r.url == url
    return reqlib


def test_normal_url(meth):
    url = "http://www.chrisheisel.com/"
    req = reqlib(url)
    result = meth(url, req)
    assert result == url


def test_sneaky_url(meth):
    url = "http://www.chrisheisel.com/?wpsrc=fol_tw"
    expected = "http://www.chrisheisel.com/"
    req = reqlib(url)
    result = meth(url, req)
    assert result == expected



import logging

import tweepy
import requests

from urlparse import urlsplit, urlunsplit


def login(consumer_key, consumer_secret, access_token, access_token_secret):
    auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
    auth.set_access_token(access_token, access_token_secret)
    api = tweepy.API(auth)
    return api


def get_since_id(api, screen_name):
    timeline = api.user_timeline(screen_name=screen_name)
    try:
        return timeline[-1].id
    except IndexError:
        return 1


def lengthen_url(url, reqlib=None):
    if reqlib is None:
        reqlib = requests
    try:
        r = reqlib.get(url)
        url = r.url
    except Exception:
        url = url

    # Ditch tracking query strings
    parts = urlsplit(url)
    parts = (parts[0], parts[1], parts[2], "", parts[4])
    return urlunsplit(parts)


def consider_status(status, cache, cache_length=604800, expand_fn=None):
    """
    Looks at a status, compares it to a cache of known urls
    and text-only Tweets. Returns either the status, or None if
    the status has been seen before.
    """
    logger = logging.getLogger("twitterdedupe.consider_status")
    if expand_fn is None:
        expand_fn = lengthen_url
    if len(status.entities['urls']) == 0:
        # Hey there's only text here
        key = str(hash("%s.%s" % (status.user.screen_name, status.text)))
        if cache.get(key) is None:
            logger.info("CACHE.MISS: %s.%s - %s" % (
                        status.user.screen_name,
                        status.id,
                        status.text
                        ))
            cache.set(key, 1, cache_length)
            return status
    else:
        # WE GOT LINKSIGN!!!
        url = status.entities['urls'][0]['expanded_url']
        expanded_url = expand_fn(url)
        key = expanded_url
        if cache.get(key) is None:
            logger.info("CACHE.MISS: %s.%s - %s" % (status.user.screen_name,
                        status.id, expanded_url))
            cache.set(key, 1, cache_length)
            return status
    return None


def get_my_unique_statuses(api, since_id, cache,
                           cache_length=604800, expand_fn=None):
    logger = logging.getLogger("twitterdedupe.get_unique_statuses")
    screen_name = api.me().screen_name
    logger.info("%s -- since %s" % (screen_name, since_id))

    stati = []

    # By using home_timeline, that means the retweeter user
    # has to be following the source(s) they want deduped
    timeline = api.home_timeline(since_id=since_id, page=1, count=100)
    for s in timeline:
        result = consider_status(s, cache, cache_length, expand_fn)
        if result is not None:
            stati.append(result)
    return stati



import pytest

from twitterdedupe.tests.helpers import (
    cache,
    MockStatus,
    MockAPI,
    nonetwork_expand_fn
)


@pytest.fixture
def meth():
    from twitterdedupe import get_my_unique_statuses
    return get_my_unique_statuses


@pytest.fixture
def one_unique_timeline():
    timeline = [
        MockStatus(
            "hitchhikers",
            44,
            "Most definitely not the question",
            "http://www.chrisheisel.com/?wpsrc=fol_tw"
        )
    ]
    return timeline


@pytest.fixture
def three_timeline():
    timeline = [
        MockStatus(
            "hitchhikers",
            44,
            "Most definitely not the question",
            "http://www.chrisheisel.com/?wpsrc=fol_tw"
        ),
        MockStatus(
            "hitchhikers",
            45,
            "Most definitely not the question but it is a dupe!",
            "http://www.chrisheisel.com/?wpsrc=fol_tw"
        ),
        MockStatus(
            "hitchhikers",
            46,
            "No links this time",
        ),
        MockStatus(
            "non-hitchhikers",
            3,
            "No links this time",
        )
    ]
    return timeline


def test_emtpy_timeline(meth, cache, nonetwork_expand_fn):
    api = MockAPI("cmheisel", [])
    result = meth(api, 1, cache, expand_fn=nonetwork_expand_fn)
    assert len(result) == 0


def test_one_unique_timeline(meth, cache,
                             nonetwork_expand_fn, one_unique_timeline):
    api = MockAPI("cmheisel", one_unique_timeline)
    result = meth(api, 1, cache, expand_fn=nonetwork_expand_fn)
    assert len(result) == 1


def test_three_timeline(meth, cache,
                        nonetwork_expand_fn, three_timeline):
    api = MockAPI("cmheisel", three_timeline)
    result = meth(api, 1, cache, expand_fn=nonetwork_expand_fn)
    assert len(result) == 3



import logging
import time

import redis
import tweepy


from twitterdedupe import login, get_my_unique_statuses, get_since_id
from twitterdedupe.caches import RedisCache


class Whitelist(logging.Filter):
    def __init__(self, *whitelist):
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record):
        return any(f.filter(record) for f in self.whitelist)


class LoggingDaemon(object):
    """
    Base daemon that calls process_status passing it each unique Tweet.
    """

    def __init__(self, env):
        self.consumer_key = env['TWITTER_CONSUMER_KEY']
        self.consumer_secret = env['TWITTER_CONSUMER_SECRET']
        self.access_token = env['TWITTER_ACCESS_TOKEN']
        self.access_token_secret = env['TWITTER_ACCESS_TOKEN_SECRET']
        self.redis_url = env['REDISTOGO_URL']
        self.screen_name = env['TWITTER_SCREEN_NAME']
        self.interval = int(env['WAIT_INTERVAL'])
        self.log_level = getattr(logging, env['LOG_LEVEL'])
        self.do_retweet = bool(int(env.get('RETWEET', False)))
        self.instance = env.get('INSTANCE')

        self.redis = redis.from_url(self.redis_url)
        self.cache = RedisCache(self.redis, "cache|%s|" % self.screen_name)

        format_str = "%(levelname)s:%(name)s::%(message)s"
        log_prefix = "%s-%s:" % (self.instance, self.screen_name)
        format_str = log_prefix + format_str

        root_handler_name = self.screen_name
        logging.basicConfig(level=self.log_level, format=format_str)
        for handler in logging.root.handlers:
            handler.addFilter(Whitelist(root_handler_name, 'twitterdedupe'))
        self.logger = logging.getLogger(root_handler_name)

        self.api = login(self.consumer_key, self.consumer_secret,
                         self.access_token, self.access_token_secret)

    def process_status(self, status):
        pass

    def run_forever(self):
        try:
            start_id = get_since_id(self.api, self.api.me().screen_name)
            while True:
                try:
                    stati = get_my_unique_statuses(self.api, since_id=start_id, cache=self.cache)
                    self.logger.info("Unique urls: %s" % len(stati))
                    try:
                        start_id = stati[0].id
                    except IndexError:
                        pass
                    [self.process_status(s) for s in stati]
                    time.sleep(self.interval)
                except tweepy.error.TweepError, e:
                    self.logger.exception(str(e))
                    self.logger.info("Waiting %s more seconds" % self.interval)
                    time.sleep(self.interval)
        except KeyboardInterrupt:
            raise
        except Exception, e:
            self.logger.exception(str(e))


class ToggleDaemon(LoggingDaemon):
    """
    Daemon that will only log by default unless it the env
    passed to it contains RETWEET=1
    """

    def process_status(self, status):
        verb = "NOOP-RETWEET"
        if self.do_retweet:
            self.api.retweet(status.id)
            verb = "RETWEET"
        self.logger.info("%s: https://twitter.com/%s/status/%s" %
                         (verb, status.user.screen_name, status.id))



import pytest

from twitterdedupe.tests.helpers import cache, MockStatus, nonetwork_expand_fn


@pytest.fixture
def meth():  # Crystal Blue Persuasion
    from twitterdedupe import consider_status
    return consider_status


@pytest.fixture
def text_status():
    status = MockStatus(
        "hitchhikers",
        42,
        "That's the question",
    )
    return status


@pytest.fixture
def link_status():
    status = MockStatus(
        "hitchhikers",
        43,
        "That's decidedly not the question",
        "http://www.chrisheisel.com/"
    )
    return status


@pytest.fixture
def repeat_link_status():
    status = MockStatus(
        "hitchhikers",
        44,
        "Most definitely not the question",
        "http://www.chrisheisel.com/"
    )
    return status


@pytest.fixture
def sneaky_link_status():
    status = MockStatus(
        "hitchhikers",
        44,
        "Most definitely not the question",
        "http://www.chrisheisel.com/?wpsrc=fol_tw"
    )
    return status


@pytest.fixture
def passthru_expand_fn():
    return lambda url: url


def test_new_text_tweet(meth, text_status, cache):
    result = meth(text_status, cache)
    assert result == text_status


def test_repeat_text_tweet(meth, text_status, cache):
    meth(text_status, cache)
    text_status.id = 999
    result = meth(text_status, cache)
    assert result is None


def test_new_link_tweet(meth, link_status, passthru_expand_fn, cache):
    result = meth(link_status, cache, expand_fn=passthru_expand_fn)
    assert result == link_status


def test_repeat_link_tweet(meth,
                           link_status,
                           repeat_link_status,
                           passthru_expand_fn,
                           cache):
    meth(link_status, cache, expand_fn=passthru_expand_fn)
    result = meth(repeat_link_status, cache, expand_fn=passthru_expand_fn)
    assert result is None


def test_sneaky_link_tweet(meth,
                           link_status,
                           sneaky_link_status,
                           nonetwork_expand_fn,
                           cache):
    meth(link_status, cache, expand_fn=nonetwork_expand_fn)
    result = meth(sneaky_link_status, cache, expand_fn=nonetwork_expand_fn)
    assert result is None



import pytest


@pytest.fixture
def nonetwork_expand_fn():
    from test_lengthen_url import reqlib
    from twitterdedupe import lengthen_url

    def wrapped_lengthen_url(url):
        return lengthen_url(url, reqlib)
    return wrapped_lengthen_url


class MemoryCache(object):
    def __init__(self, preseed=None, prefix="memtest"):
        if preseed is None:
            preseed = {}
        self.prefix = prefix
        self._data = preseed

    def _key(self, key):
        return self.prefix + key

    def keys(self):
        return self._data.keys()

    def set(self, key, value, timeout=None):
        """Set key-value in cache with given timeout (or use default one)"""
        key = self._key(key)
        self._data[key] = value

    def get(self, key):
        """Get key-value from cache"""
        key = self._key(key)
        return self._data.get(key, None)

    def flush(self, pattern='', step=1000):
        self._data = {}


@pytest.fixture
def cache():
    return MemoryCache()


class Dummy(object):
    pass


class MockStatus(object):
    def __init__(self, screen_name, status_id, text, url=None):
        self.id = status_id,
        self.screen_name = screen_name
        self.entities = {
            'urls': []
        }
        self.text = text
        if url:
            self.entities['urls'].append({'expanded_url': url})

    @property
    def user(self):
        u = Dummy()
        u.screen_name = self.screen_name
        return u


class MockAPI(object):
    def __init__(self, screen_name, home_timeline=None):
        if home_timeline is None:
            home_timeline = []
        self.screen_name = screen_name
        self._home_timeline = home_timeline

    def me(self):
        u = Dummy()
        u.screen_name = self.screen_name
        return u

    def home_timeline(self, *args, **kwargs):
        return self._home_timeline



"""
caches
From http://blog.seevl.fm/2013/11/22/simple-caching-with-redis/
"""

import cPickle


class RedisCache(object):
    """Redis cache"""

    def __init__(self, redis, prefix='cache|', timeout=60*60):
        """Create Redis cache instance using a StrictRedis connection,
        with prefix and timeout (or use default one)"""
        assert redis
        self._redis = redis
        self._prefix = prefix
        self._timeout = timeout

    def set(self, key, value, timeout=None):
        """Set key-value in cache with given timeout (or use default one)"""
        timeout = timeout or self._timeout
        key = self._prefix + key
        # Add key and define an expire timeout in a pipeline for atomicity
        self._redis.pipeline().set(key, cPickle.dumps(value)).expire(key, timeout).execute()

    def get(self, key):
        """Get key-value from cache"""
        data = self._redis.get(self._prefix + key)
        return (data and cPickle.loads(data)) or None

    def flush(self, pattern='', step=1000):
        """Flush all cache (by group of step keys for efficiency),
        or only keys matching an optional pattern"""
        keys = self._redis.keys(self._prefix + pattern + '*')
        [self._redis.delete(*keys[i:i+step]) for i in xrange(0, len(keys), step)]



from setuptools import setup, find_packages, Command

version = "0.6.1"


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys
        import subprocess
        errno = subprocess.call([sys.executable, 'runtests.py'])
        raise SystemExit(errno)


setup(
    name="twitter-dedupe",
    version=version,
    author="Chris Heisel",
    author_email="chris@heisel.org",
    description=("Python library to retweet unique links from noisy Twitter accounts."),
    long_description=open("README.md").read(),
    url="https://github.com/cmheisel/twitter-dedupe",
    zip_safe=False,
    include_package_data=True,
    scripts=['bin/daemon.py', ],
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.rst', '.html'],
    },
    packages=find_packages(),
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    install_requires=[
        "redis",
        "requests",
        "tweepy",
    ],
    cmdclass = {'test': PyTest},
)



import pytest
from mock import Mock


@pytest.fixture
def get_since_id():
    from .. import get_since_id
    return get_since_id


@pytest.fixture
def api():
    from mock import Mock
    return Mock()


def test_no_tweets(get_since_id, api):
    api.user_timeline = Mock(return_value=[])
    status_id = get_since_id(api, "cmheisel")
    assert status_id == 1


def test_one_tweet(get_since_id, api):
    mock_timeline = [
        Mock(),
    ]
    i = 111
    for s in mock_timeline:
        s.id = i
        i += 1
    api.user_timeline = Mock(return_value=mock_timeline)
    status_id = get_since_id(api, "cmheisel")
    assert status_id == 111


def test_returns_last_id(get_since_id, api):
    mock_timeline = [
        Mock(),
        Mock(),
        Mock(),
    ]
    i = 111
    for s in mock_timeline:
        s.id = i
        i += 1
    api.user_timeline = Mock(return_value=mock_timeline)
    status_id = get_since_id(api, "cmheisel")
    assert status_id == 113
